namespace dnnl {
namespace impl {
namespace gpu {
namespace ocl {
const char *ref_zero_pad_kernel[] ={ "/*******************************************************************************\n",
" * Copyright 2020 Intel Corporation\n",
" *\n",
" * Licensed under the Apache License, Version 2.0 (the \"License\");\n",
" * you may not use this file except in compliance with the License.\n",
" * You may obtain a copy of the License at\n",
" *\n",
" *     http://www.apache.org/licenses/LICENSE-2.0\n",
" *\n",
" * Unless required by applicable law or agreed to in writing, software\n",
" * distributed under the License is distributed on an \"AS IS\" BASIS,\n",
" * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n",
" * See the License for the specific language governing permissions and\n",
" * limitations under the License.\n",
" *******************************************************************************/\n",
"// This define is required before including zero_pad_struct.h to allow reuse of\n",
"// the same structure with the correct data types between OpenCL kernels and C++\n",
"// code.\n",
"#define IS_OCL_KERNEL\n",
"/*******************************************************************************\n",
" * Copyright 2020 Intel Corporation\n",
" *\n",
" * Licensed under the Apache License, Version 2.0 (the \"License\");\n",
" * you may not use this file except in compliance with the License.\n",
" * You may obtain a copy of the License at\n",
" *\n",
" *     http://www.apache.org/licenses/LICENSE-2.0\n",
" *\n",
" * Unless required by applicable law or agreed to in writing, software\n",
" * distributed under the License is distributed on an \"AS IS\" BASIS,\n",
" * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n",
" * See the License for the specific language governing permissions and\n",
" * limitations under the License.\n",
" *******************************************************************************/\n",
"#ifndef GPU_ZERO_PAD_ZERO_PAD_STRUCT_H\n",
"#define GPU_ZERO_PAD_ZERO_PAD_STRUCT_H\n",
"#define ZERO_PAD_MAX_STEP_SIZE 1024\n",
"#ifdef IS_OCL_KERNEL\n",
"#define ZERO_PAD_MASK_DATA_TYPE uchar\n",
"#else\n",
"#define ZERO_PAD_MASK_DATA_TYPE unsigned char\n",
"#endif\n",
"#define ZERO_PAD_MASK_SIZE \\\n",
"    (ZERO_PAD_MAX_STEP_SIZE / (8 * sizeof(ZERO_PAD_MASK_DATA_TYPE)))\n",
"typedef struct {\n",
"    ZERO_PAD_MASK_DATA_TYPE mask[ZERO_PAD_MASK_SIZE];\n",
"} zero_pad_mask_t;\n",
"#endif\n",
"#define DEFAULT_NELEMS_BLOCK 8\n",
"static inline void typed_ref_zero_pad(__global void *a, ulong type_size,\n",
"        ulong step_nelems, ulong nelems_block, ulong step_block, ulong nsteps,\n",
"        ulong step_size, zero_pad_mask_t step_bitmask) {\n",
"    const int i0 = get_global_id(0);\n",
"    const int istep = get_global_id(1) * step_block;\n",
"    const int iblock = get_global_id(2);\n",
"    int offset = iblock * step_size + (step_size - nsteps * step_nelems)\n",
"            + istep * step_nelems;\n",
"    const int step = 8 * sizeof(step_bitmask.mask[0]);\n",
"    // Interpret buffer differently based on type_size, this is implicitly\n",
"    // using the fact that the bit representation of 0 is the same regardless of\n",
"    // data type, e.g. int32 and f32 represent 0 the same.\n",
"    __global int *a4 = (__global int *)a;\n",
"    __global short *a2 = (__global short *)a;\n",
"    __global char *a1 = (__global char *)a;\n",
"    for (int k = 0; k < step_block; k++) {\n",
"        __attribute__((opencl_unroll_hint)) // attr:no-format\n",
"        for (int i = i0; i < step_nelems; i += nelems_block) {\n",
"            if (step_bitmask.mask[i / step] & (1 << (i % step))) {\n",
"                switch (type_size) {\n",
"                    case 4: a4[offset + i] = 0; break;\n",
"                    case 2: a2[offset + i] = 0; break;\n",
"                    case 1: a1[offset + i] = 0; break;\n",
"                }\n",
"            }\n",
"        }\n",
"        offset += step_nelems;\n",
"    }\n",
"}\n",
"static inline void sized_ref_zero_pad(__global void *a, ulong type_size,\n",
"        ulong step_nelems, ulong nelems_block, ulong step_block, ulong nsteps,\n",
"        ulong step_size, zero_pad_mask_t step_bitmask) {\n",
"    switch (type_size) {\n",
"        case 4:\n",
"            typed_ref_zero_pad((__global float *)a, 4, step_nelems,\n",
"                    nelems_block, step_block, nsteps, step_size, step_bitmask);\n",
"            break;\n",
"        case 2:\n",
"            typed_ref_zero_pad((__global float *)a, 2, step_nelems,\n",
"                    nelems_block, step_block, nsteps, step_size, step_bitmask);\n",
"            break;\n",
"        case 1:\n",
"            typed_ref_zero_pad((__global float *)a, 1, step_nelems,\n",
"                    nelems_block, step_block, nsteps, step_size, step_bitmask);\n",
"            break;\n",
"    }\n",
"}\n",
"__kernel void ref_zero_pad(__global void *a, ulong type_size, ulong step_nelems,\n",
"        ulong nelems_block, ulong step_block, ulong nsteps, ulong step_size,\n",
"        zero_pad_mask_t step_bitmask) {\n",
"    // Use a switch statement here to allow constant propagation optimizations to\n",
"    // remove switch statement from loop in typed_ref_zero_pad.\n",
"    switch (step_nelems) {\n",
"        case 16:\n",
"            sized_ref_zero_pad(a, type_size, 16, DEFAULT_NELEMS_BLOCK,\n",
"                    step_block, nsteps, step_size, step_bitmask);\n",
"            break;\n",
"        case 32:\n",
"            sized_ref_zero_pad(a, type_size, 32, DEFAULT_NELEMS_BLOCK,\n",
"                    step_block, nsteps, step_size, step_bitmask);\n",
"            break;\n",
"        case 64:\n",
"            sized_ref_zero_pad(a, type_size, 64, DEFAULT_NELEMS_BLOCK,\n",
"                    step_block, nsteps, step_size, step_bitmask);\n",
"            break;\n",
"        default:\n",
"            sized_ref_zero_pad(a, type_size, step_nelems, nelems_block,\n",
"                    step_block, nsteps, step_size, step_bitmask);\n",
"            break;\n",
"    }\n",
"}", nullptr };
}
}
}
}